﻿:Namespace Link ⍝ V 1.16
⍝ User Command to manage links/synchronisation between namespaces and file system directories
⍝ 2018 05 24 Adam: Add version number and move to DEVOPS
⍝ 2018 10 18 Adam: [16478] Move to spice by making this file monolithic
⍝ 2018 10 19 Adam: [16479] Ignore invalid names
⍝ 2018 10 19 Adam: [16480] Handle case changes
⍝ 2018 10 22 Adam: [16544,16545] Ignore invisible files and files with wrong extn
⍝ 2018 10 23 Adam: Warn user if name change in editor
⍝ 2018 10 23 Adam: Ignore inaccessible files

    ⎕IO←1 ⋄ ⎕ML←1

    :Section Main

    ∇ r←List
      r←⎕NS ⍬
      ⍝ Name, group, short description and parsing rules
      r.Name←'Link'
      r.Group←'DEVOPS'
      r.Desc←'Link ns ←→ directory'
      r.Parse←'-reset -source=ns dir both -flatten -prompt -extn= -onRead= -onWrite= -onCreate= -protect=none ns dir both -watch=none ns dir both -make=none ns dir both'
    ∇

    ∇ r←Run(cmd input)
      :Select cmd
      :Case 'Link'
          :If 0=⎕NC'⎕SE.Link'
              'Link'⎕SE.⎕NS ⎕THIS
              ⎕SE.Link.DEBUG←0
          :EndIf
          r←⎕SE.Link.UCMD.Run cmd input
      :EndSelect
    ∇

    ∇ r←level Help Cmd
      :Select Cmd
      :Case 'Link'
          r←'Link a namespace and a directory, synchronising changes.' ''
          r,←'General syntax:'('    ]',Cmd,' ns directory')
          r,←'E.g.'('    ]',Cmd,' #.utils.dates c:\devt\utils\dates')''
          r,←⊂'If only ns is provided, the only switch allowed is -reset is allowed'
          r,←'If directory is not supplied and -reset not used, a report is displayed' ''
          r,←'For more extensive documentation, see https://github.com/Dyalog/link/blob/master/README.md' ''
          r,←⊂'Modifiers:'
          r,←⊂'    -source  ={     ns|dir|both}'
          r,←⊂'              Whether to consider the ns or dir as the source (both will synchronise)'
          r,←⊂'              Defaults to "both" except when linking #, when it must be specified'
          r,←⊂'              NB: "both" will first copy from ns to dir, and then the other way'
          r,←⊂'    -watch   ={none|ns|dir|both} (after initial copying, default=both where supported, else ns)'
          r,←⊂'              Watching a dir is currently only supported under Microsoft Windows'
          r,←⊂'    -extn    =File extension considered to be APL source code (default=.dyalog)'
          r,←⊂'    -flatten  Loads everything into ns without creating sub-namespaces (default=off)'
          r,←⊂'    -prompt   Prompts user to verify all synchronisation (default=off, not recommended)'
          r,←⊂'    -reset    Removes an existing link (directory argument not required)'
          r,←⊂'    -make    ={none|ns|dir|both} (create any necessary namespaces or folders on link)'
          r,←⊂'    -dev     =Load ]link itself in developer mode (linked to source files)'
     
          r,←'' 'Switches defining hooks to pre-process synchronisation' ''
          r,←⊂'    -onRead   Function to call when an external change is detected'
          r,←⊂'              The fn will receive (change-type file-name target-ns-name)'
          r,←⊂'              Return 1 if link should perform own processing,'
          r,←⊂'              0 to signal that hook has performed all necessary actions.'
          r,←⊂'              NB: also called for files which do not have the specified extension.'
          r,←⊂''
          r,←⊂'    -onWrite  Function to call when an internal change is detected'
          r,←⊂'              The fn will receive (namespace object-name old-name name-class source-as-vtv linked-file-name)'
          r,←⊂'              Return 1 if link should perform own processing,'
          r,←⊂'              0 to signal that hook has performed all necessary actions.'
          r,←'' 'Currently unsupported switches:' ''
          r,←⊂'    -protect ={none|ns|dir|both}'
          r,←⊂'              Whether to protect from overwriting when copying initial source'
      :EndSelect
    ∇

    ∇ {linked}←where Fix src;link;nss;files;file;src;nc;nosrc;fsw;old;ns;name;⎕TRAP;oldname;z
⍝ Fix a function/operator or script, preserving any existing source files
⍝   Used internally by EditorFix "afterfix" processing
⍝   May be called by other tools providing the source in ⎕NR/⎕SRC format on the right
⍝   One day, the left argument of (ns name) may be inferred, for now it must be provided
     
⍝ Returns 1 if a link was found for the name, else 0
⍝   NB: if 0 is returned, no ⎕FX/⎕FIX was done
     
      :If DEBUG=2
          ⎕TRAP←0 'S' ⋄ ∘∘∘
      :EndIf
     
      nosrc←0=≢src
      (ns name oldname)←3↑where,⊂''
      z←ns←⍬⍴ns
     
      (linked link)←0 ⍬
     
      :If 0≠⎕NC'⎕SE.Link.Links'
      :AndIf 0≠≢⎕SE.Link.Links
          nss←⍎¨⎕SE.Link.Links.ns
          :Repeat
              :If ~linked←z∊nss
                  linked←(z←z.##)∊nss
              :EndIf
          :Until linked∨z∊# ⎕SE
          :If linked
              link←(nss⍳z)⊃⎕SE.Link.Links
          :EndIf
     
      :EndIf
     
      :If linked
          :If 0<≢file←4⊃Utils.GetLinkInfo ns name ⍝ Already saved in a file
          :AndIf name≡oldname                  ⍝ Function not renamed
     
          :ElseIf ~link.flatten                ⍝ Not flattened
     
              file←(≢link.ns)↓⍕ns              ⍝ Add sub.namespace structure
              ((file='.')/file)←'/'            ⍝ Convert dots to /
              file←link.dir,file               ⍝ Add link directory
              file,←'/',name,link.extn         ⍝ And the object name
     
          :Else                                ⍝ New name in "flattened" link
              :If 3=link.⎕NC'onCreate'         ⍝ User exit to determine file name?
                  →(0=≢file←(link⍎onCreate)ns name)⍴0 ⍝ Return file name or '' to abort
              :Else
                  ⎕←'Fix: unable to determine file name for "',name'"'
                          ⍝ /// could use folder browser under Windows
              :EndIf
          :EndIf
     
          nc←⍬⍴ns.⎕NC name
          :If nosrc                       ⍝ Source not provided - find it
              :Select nc
              :Case 2
                  src←ns⍎name
              :CaseList 3 4
                  src←ns.⎕NR name
              :Case 9
                  src←⎕SRC ns⍎name
              :Else
                  ⎕←'Fix: unable to handle ',(⍕ns),'.',name,' (⎕NC =',(⍕nc),')'
                  →0
              :EndSelect
          :EndIf
     
          :If 2=⎕NC'_rename_temp' ⍝ // workaround for editor overwriting old file on rename
          :AndIf (name oldname)≡(1⊃_rename_temp)[6 5]
              (3⊃_rename_temp)⎕NPUT(2⊃_rename_temp)1 ⍝ Restore old file contents
              ⎕EX'_rename_temp' ⍝ Now leesten carefully, for I will do zis only vunce
          :EndIf
     
          :If 3=⎕NC link.onWrite
              →((⍎link.onWrite)ns name oldname nc src file)↓0
              →(nc=2)⍴0 ⍝ We don't do variables
          :EndIf
     
          :Trap 0
              :If fsw←0≠⎕NC⊂'link.fsw.Object.EnableRaisingEvents'
                  old←link.fsw.Object.EnableRaisingEvents
                  link.fsw.Object.EnableRaisingEvents←0
              :EndIf
              3 ⎕MKDIR 1⊃⎕NPARTS file ⍝ make sure folder is there
              (⊂src)⎕NPUT file 1
          :Else
              old←0
              ⎕←'Fix: Unable to write "',name,'" to file ',file
              ⎕TRAP←0 'S' ⋄ ∘∘∘
              ⎕←'     '⎕DMX
          :EndTrap
          :If fsw ⋄ link.fsw.Object.EnableRaisingEvents←old ⋄ :EndIf
     
          :Trap 0
              2 ns.⎕FIX'file://',file
          :Else
              ⎕←'Fix: Unable to re-fix "',name,'" from file ',file
              ⎕TRAP←0 'S' ⋄ ∘∘∘
              ⎕←'     '⎕DMX
          :EndTrap
      :EndIf
    ∇
    ∇ {r}←Expunge name;caller;src
⍝ Use in place of ⎕EX for linked names - will expunge source files where relevant
     
      :If DEBUG=2
          ⎕TRAP←0 'S' ⋄ ∘∘∘
      :EndIf
     
      caller←⊃⎕RSI
      src←Utils.GetLinkInfo caller name
      r←caller.⎕EX name
     
      :If (0≢4⊃src)∧0=≢4⊃Utils.GetLinkInfo caller name ⍝ There was source and now there is none
          ⎕NDELETE 4⊃src
      :EndIf
    ∇
    ∇ {r}←Notify args;type;path;oldpath;linked;link;i;folder;name;extn;U;nsname;ns;effect;z;affected;newname;oldname;resp;dispPath;dispNs;cap;text;⎕TRAP;links;Check;above;regex
⍝ Notify Link system that an external file has changed
⍝ Usually called by FileSystemWatcher
     
      :If DEBUG>0
          ⎕←args
      :AndIf DEBUG>1
          ⎕TRAP←0 'S' ⋄ ∘∘∘
      :EndIf
     
      Check←→0⍴⍨¯1∊∘⎕NC'/'∘≠⊆⊢ ⍝ leave early if any bad name
     
      r←⍬
      U←Utils
     
      (type path oldpath)←{⍵,(≢⍵)↓'' '' ''},⊆args
      (path oldpath)←{(0≠≢⍵)/U.Normalise ⍵}¨path oldpath
     
      linked←{0::⍬ ⋄ Utils.Normalise¨⎕SE.Link.Links.dir}⍬
      →((≢linked)<i←1⍳⍨linked≡¨(≢¨linked)↑¨⊂path)⍴0 ⍝ not linked => done
      link←i⊃⎕SE.Link.Links
     
      (folder name extn)←⎕NPARTS path
      Check name ⍝ invalid item name
     
      regex←'^' '\*' '\?' '$'⎕R'^' '.*' '.' '$'⊢link.extn~'.' ⍝ convert glob to regex
      →0~regex ⎕S 2⊢extn~'.' ⍝ check extn
     
      folder←¯1↓folder ⍝ drop trailing /
     
      nsname←(1+≢link.dir)↓folder
      Check nsname/⍨link.flatten
      ((nsname='/')/nsname)←'.'
     
      ⍝ check visibility
      Ancestry←(⊣,'/',⊢)\'/'(≠⊆⊢)∘⊃1∘⎕NPARTS
      above←≢Ancestry link.(dir↓⍨-'/'=⊃⌽dir)
      :Trap 0 ⍝ no access
          →0/⍨∨/6 ⎕NINFO(above↓Ancestry name),⊂path
      :Else
          →0
      :EndTrap
     
      nsname←link.ns,(~link.flatten)/((0≠≢nsname)/'.'),nsname ⍝ target namespace
     
      :If 3=⎕NC link.onRead ⍝ user exit defined?
          :Trap 0
              →((link⍎link.onRead)type path nsname)↓0 ⍝ return 1 to continue else 0
              :If ~(type≡'deleted')∨(U.IsDir path)∨extn≡link.extn
                  Log'Ignoring due to extension: ',type,' of ',path
                  →0
              :EndIf
          :Else
              ⎕←'*** onRead callback ',link.onRead,' failed: ',⎕DMX.EM
              ∘∘∘
          :EndTrap
      :EndIf
     
      :If 0=≢ns←U.GetRefTo nsname ⍝ We're going to need a ref to it
          Log'Unable to create namespace ',nsname
          →0 ⍝ might as well give up now
      :EndIf
     
      effect←0
     
      :Select type
     
      :Case 'created'               ⍝ A new file
          effect←0                  ⍝ may not have an effect on existing objects
          :If 0≠≢z←U.GetName path     ⍝ try to extract the name
              name←z
          :EndIf
          :If 0≠≢z←4⊃U.GetLinkInfo ns name  ⍝ Redefines existing object with a source file
          :AndIf ⎕NEXISTS z       ⍝ ... and that file exists
              affected←(⍕ns),'.',name
              Log'(created): ignoring new file ',path
              Log'           attempts to redefine ',affected,' which is linked to ',z
              →0
          :EndIf
     
      :Case 'changed'               ⍝ Update to existing file?
          :If 0≠≢z←4⊃U.GetLinkInfo ns name      ⍝ name already linked
          :AndIf name≡U.GetName path            ⍝ name matches file contents
          :AndIf (U.Normalise path)≡U.Normalise 4⊃z  ⍝ to the same file name
              effect←1 ⋄ affected←(⍕ns),name
              :If (7⊃z)≡U.GetCheckSum path
                  Log'(change) ignored, checksum unchanged: ',path
                  →0
              :Else
                  Log'(change): ',affected,' updated in ',path
              :EndIf    ⍝ ... but the file did not change
     
          :Else ⍝ apparent change to name with no current link
              :If 0≠≢z←U.GetFileInfo path ⍝ Paranoia: check if file defined another name
                  :If 1≠≢z
                      ⎕←'WARNING: user changed name of edited item - ]Link lost track!'
                      ⎕←'   Best course of action is:'
                      ⎕←'      ]Link ',(⍕link.ns),' -reset'
                      ⎕←'      ]Link ',(⍕link.ns),' ',link.dir
                      →0
                  :EndIf ⍝ Multiple names
                  z←⊃z
              :AndIf 0≠(2⊃z).⎕NC⍕1⊃z     ⍝ If the old name exists
                  (2⊃z).⎕EX⍕1⊃z          ⍝ Expunge it
              :EndIf
     
              :If U.IsDir path
                  Log'(changed) ignored "change" to directory: ',path
                  →0
              :EndIf
     
              :If name≢newname←U.GetName path             ⍝ repeated in create
              :AndIf 0≠≢⊃z←U.GetLinkInfo ns newname       ⍝ name inside is already linked
                  affected←(⍕ns),'.',name
                  Log'(created): ignoring changed file ',path
                  Log'           attempts to redefine ',affected,' which is linked to ',4⊃z
                  →0
              :EndIf
     
              :If 0≠≢z←U.GetFileInfo path                ⍝ Try to find an "affected" object
                  effect←1 ⋄ affected←(⍕ns),'.',name
                  Log'(changed): ',affected,' was previously linked to ',path
              :ElseIf 0≠≢newname
                  effect←1 ⋄ affected←(⍕ns),'.',newname   ⍝ So just accept it
                  Log'(changed): fixing ',affected,' from ',path
              :Else
                  Log'(changed): No APL object found in ',path
                  →0
              :EndIf
     
          :EndIf
     
      :Case 'renamed'
          :If U.IsDir path                              ⍝ of a folder?
          :AndIf 0=ns.⎕NC name                          ⍝ and new name is free and valid
              :If 9=ns.⎕NC oldname←∊1↓⎕NPARTS oldpath   ⍝ and old name looks like an existing namespace
                  ns⍎name,'←',oldname                   ⍝ copy it
                  (⍎(⍕ns),'.',name).⎕DF(⍕ns),'.',name   ⍝ give it a nice name
                  ns.⎕EX oldname                        ⍝ expunge old
                  links←U.Normalise¨4⊃¨5177⌶⍬
                  links←(((≢oldpath)↑¨links)∊⊂oldpath)/links
                  links←(⊂path),¨(≢oldpath)↓¨links
                  :For link :In links
                      Notify'changed'link
                  :EndFor
     
                  Log'NS rename ',oldname,' → ',(⍕ns),'.',name
              :Else                                    ⍝ no old ns
                  Log'NS rename ',oldname,' (not found) → ',name
              :EndIf
              →0
          :EndIf
     
          oldname←2⊃⎕NPARTS oldpath                    ⍝ rename - not of a folder
          :If 0≠≢⊃z←U.GetLinkInfo ns oldname           ⍝ old name *is* linked to this file
              effect←1 ⋄ affected←(⍕ns),'.',name
              Log'(rename) ',(⍕ns),'.',oldname,' → ',affected
     
          :Else ⍝ old name not linked to old file
              :If 0≠≢z←U.GetFileInfo path
                  affected←(⍕2⊃⊃z),'.',⊃⊃z
                  Log'(rename): ',affected,' was previously linked to ',path
              :Else
                  :If 0≠≢name←U.GetName path
                      :If 0≠≢⊃z←U.GetLinkInfo ns name ⍝ but name inside is already linked
                      :AndIf ⎕NEXISTS 4⊃z             ⍝ to a file which exists
                          affected←(⍕ns),'.',name
                          Log'(created): ignoring changed file ',path
                          Log'           attempts to redefine ',affected,' which is linked to ',4⊃z
                          →0
                      :Else
                          effect←1 ⋄ affected←(⍕ns),'.',name
                          Log'(rename): re-definition of ',affected,' from ',path
                      :EndIf
                  :Else
                      Log'(rename): Unable to find changed object corresponding to ',path
                  :EndIf
              :EndIf
          :EndIf
     
      :Case 'deleted'
          :If 0≠≢U.GetLinkInfo ns name   ⍝ existing name is linked to this file
          :OrIf 9.1=ns.⎕NC⊂name          ⍝ it identifies a namespace
              effect←1 ⋄ affected←(⍕ns),'.',name
              Log'(delete) of ',affected
          :ElseIf 0≠≢z←U.FindLinkInfo path
              effect←1 ⋄ affected←2⊃z
              Log'(delete): ',affected,' was previously linked to ',path
          :Else
              Log'(delete): Unable to find changed object corresponding to ',path
          :EndIf
     
      :EndSelect
     
      :If effect∨'created'≡type
     
          :If ~resp←link.quiet
     
              dispPath←U.WinSlash Strip path
              dispNs←U.WinSlash⍕ns ⍝ / → \ on Windows
              cap←'Namespace-Directory Synchroniser'
     
              :If type≡'renamed'
                  text←⊂'"',(U.WinSlash oldpath),'" has been renamed to "',(U.WinSlash path),'". Would you like to synchronise?'
              :Else
                  text←⊂'"',(U.WinSlash path),'" has been ',type,'. Would you like to synchronise?'
              :EndIf
     
              :Select type
     
              :Case 'created'
                  text,←''('If you choose Yes, the contents of the file will be established in "',dispNs,'".')
                  text,←''('If you choose No, "',dispNs,'" will be out of sync with "',dispPath,'".')
              :Case 'changed'
                  text,←''('If you choose Yes, "',affected,'" will be updated.')
                  text,←''('If you choose No, "',dispNs,'" will be out of sync with "',dispPath,'".')
              :Case 'renamed'
                  text,←''('If you choose Yes, "',affected,'" will be re-mapped.')
                  text,←''('If you choose No, "',dispNs,'" will be mapped to a file which may no longer exist.')
              :Case 'deleted'
                  text,←''('If you choose Yes, "',affected,'" will be DELETED!')
                  text,←''('If you choose No, "',dispNs,'" will contain items that are not on in the "',dispPath,'".')
              :EndSelect
     
              resp←⎕SE.SALTUtils.msgBox cap text'Query'
          :EndIf
     
          :If 1=resp
              :If (⊂type)∊'renamed' 'deleted'
                  ⎕EX affected
              :EndIf
     
              :If type≢'deleted'
                  :Trap 0
                      :If U.IsDir path
                          name ns.⎕NS''   ⍝ create namespace
                      :Else               ⍝ file
                          2 ns.⎕FIX'file://',path
                      :EndIf
                  :Case 22
                      :If ~'tmp___'≡¯6↑path ⍝ PyCharm up to no good again
                          ⎕←'(',type,') reported to unreadable file ',path
                      :EndIf
                  :Else
                      ⎕←'Unable to fix file ',path
                      ⎕←'   '⎕DMX
                  :EndTrap
              :EndIf
          :EndIf
      :EndIf
    ∇
    ∇ Log x
      :If DEBUG≠0
          ⎕←x
      :EndIf
    ∇
    :EndSection
    :Namespace UCMD
        ∇ r←Run(Cmd Input);U;argcount;container;dir;doDir;doNs;extn;flatten;fsw;infail;isWin;ix;l;linkDef;m;make;ns;onRead;onWrite;outfail;p;protect;quiet;reset;source;w;watch;z
          :Select Cmd
          :Case 'Link'
              U←##.Utils
         
              :If 0=⎕NC'⎕SE.Link.Links'    ⍝ Check existence of our data structure
                  ⎕SE.Link.Links←⍬         ⍝ Start with no links
              ⍝ Link namespaces will record switch values
              ⍝    and also contain fsw, the FileSystemWatcher object
              :EndIf
         
              (source protect make)←⊂¨'both'∘Input.Switch¨'source' 'protect' 'make'
              isWin←'Windows'≡7↑⊃'.'⎕WG'APLVersion'
              watch←⊂((1+isWin)⊃'ns' 'both')∘Input.Switch'watch' ⍝ Default to "both" under Windows, else "ns"
              protect←⊂'none'Input.Switch'protect' ⍝ /// while protect is not supported
              extn←'.dyalog'Input.Switch'extn'     ⍝ File extension to use
              flatten←0 Input.Switch'flatten'      ⍝ Whether to flatten the external folder structure
              quiet←~0 Input.Switch'prompt'        ⍝ do not prompt on external changes?
              reset←0 Input.Switch'reset'          ⍝ Whether to flatten the external folder structure
         
              (onRead onWrite)←''∘Input.Switch¨'onRead' 'onWrite'
         
              :If ~isWin
              :AndIf (⊂watch)∊'both' 'dir'
                  →0⊣r←'Watching directories is only supported under Microsoft Windows'
              :EndIf
         
              :If protect≢⊂'none'
                  →0⊣r←'-protect not yet supported'
              :EndIf
         
              ns←⊃Input.Arguments
              :If 0≠argcount←≢Input.Arguments    ⍝ ns provided
                  container←⎕RSI⊃⍨⎕SI⍳⊂'UCMD'    ⍝ ns is relative to where we were called from
         
                  :If (ns≡,'#')∨9=container.⎕NC ns
                      ns←container⍎ns
                  :ElseIf make∊'ns' 'both'
                      :Trap 0
                          z←(⊂'#.')~⍨{((⍸⍵='.')↑¨⊂⍵),⊂⍵}ns
                          z←(0=container.⎕NC z)/z
                          z container.⎕NS¨⊂''
                          ns←container⍎ns
                      :Else
                          →0⊣r←'Unable to create: ',ns
                      :EndTrap
                  :Else
                      →0⊣r←ns,' is not a namespace'
                  :EndIf
              :EndIf
         
              l←⎕SE.Link.Links
              :If 1=argcount     ⍝ only ns provided
                  :If 0=≢l
                  :OrIf 1≠⍴l←((⍎¨l.ns)∊ns)/l          ⍝ reduce l
                      →0⊣r←'No link defined for namespace ',ns
                  :EndIf
              :EndIf
         
              :Select argcount
              :CaseList 0 1 ⍝ ns only provided, or nothing: Report or Reset
                  :If reset
                      :If argcount=0
                          :If 0=⍴⎕SE.Link.Links ⋄ r←'Nothing to reset'
                          :Else ⋄ r←⍪ResetLink¨l
                          :EndIf
                      :ElseIf argcount≠1 ⋄ r←'-reset requires selection of a namespace'
                      :Else ⋄ r←ResetLink ⍬⍴l
                      :EndIf
                      →0
                  :Else
                      →0⊣r←ReportLinks l
                  :EndIf
              :Case 2       ⍝ ns and dir: Create a link
                  :If 0<≢l
                  :AndIf (⊂ns)∊⍎¨l.ns
                      →0⊣r←↑('A link already exists for ',(⍕ns),'- please use:')('      ]link ',(⍕ns),' -reset')
                  :EndIf
         
                  :If ~⎕NEXISTS dir←U.Normalise 2⊃Input.Arguments
                      :If make∊'dir' 'both'
                          3 ⎕MKDIR dir
                      :Else
                          →0⊣r←'Directory not found: ',dir
                      :EndIf
                  :EndIf
         
                  :If 0≡Input.Switch'source'
                  :AndIf ∧/0≠(≢ns.⎕NL⍳10),≢⊃(⎕NINFO⍠1)dir,'/*'
                      →0⊣r←'-source must be specified when linking a non-empty namespace to a non-empty folder'
                  :EndIf
         
              :Else
                  'requires one or two arguments: [ns[dir]]'⎕SIGNAL 11
              :EndSelect
         
              '-flatten requires -source=dir'⎕SIGNAL(flatten∧source≢⊂'dir')/11 ⍝ cannot export and flatten
         
              doNs←'ns' 'both'
              doDir←'dir' 'both'
              r←⊂(⍕ns),' linked to ',dir
              infail←outfail←fsw←⍬
              ix←'[',(⍕1+≢⎕SE.Link.Links),']'
         
              :If source∊doNs
                  (p m)←protect make∊¨⊂doDir
                  w←watch∊doNs
                  :If 0≠≢outfail←ns Export dir p w m extn flatten quiet
                      r,←('Failed to export ',(⍕≢outfail),' file(s) - see:')('      ⎕SE.Link.Links',ix,'.outfail')
                  :EndIf
              :EndIf
         
              :If source∊doDir
                  (p m)←protect make∊¨⊂doNs
                  w←watch∊doDir
                  (infail fsw)←ns Import dir p w m extn flatten quiet
                  :If 0≠≢infail
                      r,←('Failed to import ',(⍕≢infail),' file(s) - see:')('      ⎕SE.Link.Links',ix,'.infail')
                  :EndIf
              :EndIf
         
              :If 0=≢r
                  r←⍬
              :EndIf
         
              (linkDef←⎕SE.⎕NS ⍬).⎕DF'[Link]'
              linkDef.(ns dir extn flatten protect watch fsw infail outfail)←((⍕ns)dir extn flatten protect watch fsw infail outfail)
              linkDef.(onWrite onRead quiet)←onWrite onRead quiet
              ⎕SE.Link.Links,←linkDef
         
              r←⍕⍪r
          :EndSelect
        ∇

        ∇ r←ResetLink link;m;ol;i;name;ns;n
     ⍝ Reset a link
         
          ol←objectLinks ⍬
          m←⍸(2⊃¨ol)within⍎link.ns
          m←m[⍋3⊃¨ol[m]] ⍝ members first, classes last
         
          :For i :In m
              (name ns)←2↑i⊃ol
              {}5178(ns.⌶)⍕name
          :EndFor
         
          :If 9=⎕NC'link.fsw'
              link.fsw.Object.EnableRaisingEvents←0
          :EndIf
         
          ⎕SE.Link.Links~←link
          r←'Link from ',(⍕link.ns),' to ',(⍕link.dir),' removed (',(⍕≢m),' names affected)'
        ∇

        ∇ r←ReportLinks links;count
          :If 0=≢links
              →0⊣⊂r←'No source links defined'
          :EndIf
         
          count←+/¨((2⊃¨objectLinks ⍬)within)¨⍎¨l.ns
         
          r←⊂'Namespace' 'Directory' 'Names'⍪(↑l.(ns dir)),count
          r←⍕⍪r
        ∇

          within←{           ⍝ nss within parent
              ⍺←0            ⍝ first level call
              r←⍺∨⍺⍺∊⍵       ⍝ well, is it?
              ∧/r:r          ⍝ they all are
              m←⍺⍺∊# ⎕SE
              ∧/m:r          ⍝ we reached #
              i←⍸~m
              r[i]←r[i](⍺⍺[i].## ∇∇)⍵  ⍝ try parents
              r
          }

        objectLinks←5177⌶

          Import←{ ⍝ Load items from files in folders
              ⍺←#                                      ⍝ default is root
              target←⍺                                 ⍝ left arg is namespace to put items in
              (source protect watch make extn flatten quiet)←⍵ ⍝ right arg is where to load from, check for existing, break connection
              detach←~watch                            ⍝ if we don't need to watch, simply detach
     ⍝disperse←1
         
              Slash←∊∘'/\'                             ⍝ Mark slashes
              Parts←~∘Slash⊆⊢                          ⍝ Path parts
         
              source↓⍨←-Slash⊃⌽source                  ⍝ Remove slash from end
              cands←⊃⎕NINFO⍠1⊢source,'*'               ⍝ normalised source candidates
              candLengths←≢¨cands                      ⍝ their lengths
         
              3::(⍵~0)⍬                               ⍝ if unlisted
         
              source←cands⊃⍨candLengths⍳≢source        ⍝ first length-match
              name←⊃⌽Parts source                      ⍝ name without path
     ⍝~disperse:source protect detach 1 ∇⍨⍎name ⍺.⎕NS ⍬ ⍝ adjust level and call again
         
              File←detach{⍺⍺:⊃⎕NGET ⍵ 1 ⋄ 'file://',⍵} ⍝ Prepare file by getting source or prepending prefix
         
              2≡1 ⎕NINFO source:watch{ ⍝ if it is a file, just fix it
                  _←2 target.⎕FIX File ⍵
                  ~⍺:⍬ ⍬
                  d f e←⎕NPARTS ⍵
                  ⍬(##.FileSystemWatcher.Watch d(f,e))
              }source
              source,←'/'                              ⍝ append missing slash
              list←0 1 6 ⎕NINFO⍠1⍠'Recurse' 1⊢source,'*' ⍝ recursive listing of everything
         
              Where←⌿⍨                                 ⍝ Filter as function
              Begins←⊃⍷                                ⍝ ⍵ starts with ⍺
         
              hideBit←1=3⊃list                         ⍝ mask for hidden items
              hidden←⊃list Where¨⊂hideBit              ⍝ list of hidden (files and) folders
              hidden,¨←'/'                             ⍝ protect agains similarly named items
              list←list Where¨⊂~hideBit                ⍝ keep only visible items
              list←list Where¨⊂~∨⌿hidden∘.Begins⊃list  ⍝ filter away things that come below hidden things
              list↓⍨←¯1                                ⍝ remove hideBit column
         
              RightExtn←extn≡(-≢extn)↑⊢                ⍝ Ends with specified extension?
         
              Files←⊢Where RightExtn¨                  ⍝ Those that end with specified extension
         
              files←Files⊃list                         ⍝ first column is filename
              dirs←⊃Where/1=@2⊢list                    ⍝ second column has 1 for dirs
              dirs,¨←'/'                               ⍝ give the dirs trailing slashes
              list←dirs⍪files                          ⍝ we need to process dirs before files
              order←⍋1114112@(47=⊢)↑⎕UCS¨list          ⍝ slashes after all chars
              list←order⊃¨⊂list                        ⍝ sort the list
         
              Path←⊢↓⍨∘-'/'⍳⍨⌽                         ⍝ Until last slash
              RemDir←(≢source)∘↓                       ⍝ Only path beginning at target source
         
              paths←∪RemDir∘Path¨list                  ⍝ paths of all unique items
              paths~←⊂''                               ⍝ remove empty
         
              Nss←,\'.',¨1↓Parts                       ⍝ Convert file path to ns path
              DotSlash←'.'@Slash                       ⍝ Convert dots to slashes
         
              _←{⍵:⍬                                   ⍝ If we are not flattening, create nss
                  nss←DotSlash¨paths                   ⍝ all necessary namespaces
                  nss←(¯1≠⎕NC nss)/nss                 ⍝ ignore invalid names
                  NsExpr←'''',,∘'''⎕NS⍬'               ⍝ Expression to create namespace
                  target.⍎∘NsExpr¨(~flatten)/nss       ⍝ create namespaces in target
              }flatten
         
              PadThis←⊢,'⎕THIS'/⍨0=≢                   ⍝ Default path to here
              Target←'.'@Slash∘PadThis∘((~flatten)∘/)Path∘RemDir ⍝ Determine target path
         
              FixThere←{ ⍝ Fix the file ⍵ in the namespace ⍺
                  t←(1+flatten)⊃(⍺.⍎Target ⍵)⍺         ⍝ final target
                  2 t.⎕FIX File ⍵ ⍝ do it
              }
         
              TryFixThere←{      ⍝ Try to fix file there
                  ⍵∧.=' ':⍬      ⍝ Called by each on prototype: ignore
                  6 11::⍵        ⍝ return name on failure
                  0⊣⍺ FixThere ⍵ ⍝ do it and return 0 on success
              }
         
              r←0~⍨target∘TryFixThere¨Files list ⍝ try it and return only failures
     ⍝ return (list of failures) and (watcher reference or zilde)
              r({⍵:##.FileSystemWatcher.Watch(¯1↓source)(,'*') ⋄ ⍬}watch)
          }

          Export←{ ⍝ Write items to file in folders
         
              ⍺←#                           ⍝ default is root
              source←⍺                      ⍝ left arg is namespace get items from
              (dir protect watch make extn flatten quiet)←⍵ ⍝ right arg is where to store files and whether to protect
              overwrite←~protect            ⍝ make positive
              dir,←'/'/⍨~'/\'∊⍨⊃⌽dir        ⍝ append trailing slash if missing
              dir←⊃1 ⎕NPARTS dir            ⍝ normalise ─ '' means current dir
         
              Enclose←⊢ ⍝ ,¨@1,⊣            ⍝ Add ∇s to ⎕NR
              Path←dir,'/'⊣@('.'=⊢)⊢        ⍝ Ns path → full path
              Path←Path,'/'extn⊃⍨⊣      ⍝ Append appropriate tail (⍺=1:dir; ⍺=2:file)
              Into←overwrite{ ⍝ Put vtv into file
                  22::⍵                  ⍝ no file access
                  _←3 ⎕MKDIR⊃1 ⎕NPARTS ⍵ ⍝ create dir if needed
                  0⊣⍵ ⍺⍺ ⎕NPUT⍨⊂⍺        ⍝ overwrite
              }
              AbsDelRep←'∇'Enclose ⎕NR  ⍝ Put ∇s on ⎕NR
              Last←⊢↑⍨1-'.'⍳⍨⌽          ⍝ Item name without ns path
              AbsSaveFn←AbsDelRep Into 2 Path Last ⍝ Save single fn into target
         
              ''≡0⍴⍺:0⍴⊂0⍴AbsSaveFn source ⍝ single fn/op
         
              OnEach←{ ⍝ ¨ without prototype call on empty
                  0∊⍴⍵:⍬ ⍝ if empty return empty
                  ⍺←⊢    ⍝ ambivalent
                  ⍺ ⍺⍺¨⍵ ⍝ call
              }
         
              SaveNs←{ ⍝ Save namespace whether scripted or not
                  0=≡⍵:0                ⍝ prototype means
                  16::0 ⍵               ⍝ problem: no script
                  src←⎕SRC source.⍎⍵    ⍝ get source from its location
                  22::⍵ 0               ⍝ failed: no file access
                  0 0⊣src Into 2∘Path ⍵ ⍝ save as namespace and report OK
              }
         
              nss←source.⎕NL ¯9.1 ¯9.4 ¯9.5              ⍝ namespaces
         
              No0←~∘0                                    ⍝ Remove fills
              DelRep←'∇'Enclose source.⎕NR               ⍝ ⎕NR with ∇s there
              SaveFn←DelRep Into 2∘Path                  ⍝ Save function there
         
              failed unscripted←No0¨↓⍉↑SaveNs OnEach nss ⍝ go through them
              paths←1∘Path¨unscripted                    ⍝ needed dirs
              _←3 ⎕MKDIR paths                           ⍝ make them if they don't exist
              ok←⎕NEXISTS paths                          ⍝ did it work?
              fns←source.⎕NL-3.1 3.2 4.1 4.2             ⍝ tradfn, dfn, tradop, dop
              failed,←No0 SaveFn OnEach fns              ⍝ try to save the functions
              failed,←paths/⍨~ok                         ⍝ also remember failed paths for nss
              paths/⍨←ok                                 ⍝ the good paths
              unscripted/⍨←ok                            ⍝ and their namespaces
         
              Ref←source.⍎⊣                              ⍝ Translate text larg to ref
              Recurse←Ref ∇ protect watch make extn flatten quiet,⍨⊢∘⊂ ⍝ How to call into subs
         
              failed,⊃,/unscripted Recurse OnEach paths  ⍝ recurse and collect failures
          }
    :EndNamespace
    :Namespace FileSystemWatcher

        (⎕IO ⎕ML)←1 1

        ∇ WatchEvent(obj args);nargs
     ⍝ Callback for System.IO.FileSystemWatcher instance
  ⍝     Passes info on to ⎕SE.Link.Notify for processing
         
          {}2501⌶0 ⍝ kill thread on exit
         
          nargs←⊂819⌶⍕args.ChangeType             ⍝ type: created|changed|deleted|renamed
          nargs,←⊂args.FullPath
          :If 0≠⎕NC⊂'args.OldFullPath'
              nargs,←⊂args.OldFullPath
          :EndIf
         
          {}⎕SE.Link.Notify nargs
        ∇

        ∇ r←Watch args;⎕USING;path;filter;watcher
     ⍝ args: path filter
         
          ⎕USING←',System.dll'
          watcher←⎕NEW System.IO.FileSystemWatcher
          watcher.(Path Filter)←args
          watcher.(onChanged onCreated onDeleted onRenamed)←⊂'WatchEvent'
          watcher.IncludeSubdirectories←1
          watcher.EnableRaisingEvents←1
         
          r←⎕NEW Disposable watcher ⍝ wrap a destructor around it
        ∇

        :Class Disposable

            :Field Public Object

            ∇ make ref
              :Access Public
              :Implements Constructor
              Object←ref
            ∇

            ∇ do_dispose ref
              :Trap 0
                  ref.Dispose
              :EndTrap
            ∇

            ∇ dispose;tid
              :Implements Destructor
              Object.EnableRaisingEvents←0
              tid←do_dispose&Object
            ∇

        :EndClass

    :EndNamespace
    :Namespace Utils
    ⍝ Utilities shared by Link infrastructure

        Normalise←∊1⎕NPARTS⊢

        ∇ r←IsWindows
          r←⎕SE.SALTUtils.WIN
        ∇

        Strip←{⍵↓⍨-'/'⍳⍨⌽⍵}
        Shortest←{⍵⊃⍨(⊢⍳⌊/)≢¨⍵}
        WinSlash←'\'@(=∘'/')⍣IsWindows
    Combine←{(⍕⍺),⍨(326≠⎕DR ⍺)/'.',⍨⍕⍵}/
        IsDir←{22::0 ⋄ 1=1 ⎕NINFO ⍵}

        lc←819⌶ ⍝ Lowercase

        GetCheckSum←{2 ⎕NQ #'GetBuildID' ⍵}

        ∇ r←GetName file;code;ns
    ⍝ Attempt to determine the name which will be defined by a file
         
          ns←(⎕NS'').⎕NS r←'' ⍝ Two levels of nesting to allow safe fixing of :Namespace/##.⎕IO←0/:EndNamespace
          :Trap 0
              2 ns.⎕FIX'file://',file
              :If 1=≢r←ns.⎕NL-⍳10 ⋄ r←⊃r
              :Else ⋄ r←''
              :EndIf
          :EndTrap
        ∇

        ∇ r←GetLinkInfo(ns name)
    ⍝ Return link info or ⍬
          r←5179(ns.⌶)name
        ∇

        ∇ r←GetFileInfo file
    ⍝ Return links to file
          r←5174⌶file
        ∇

        ∇ r←GetRefTo nsname;name
     ⍝ Get a reference to a container namespace (or ⍬ if that is not possible)
         
          :If ~(⊂nsname)∊'⎕SE'(,'#')
              r←⍬ ⍝ pessimism in action
              →(¯1=⎕NC nsname)⍴0 ⍝ Not a valid name, not worth trying
         
              :If ~(nsname≡,'#')∨9=⎕NC nsname ⍝ if it isn't already there
                  :Trap 0
                      :For name :In (¯1+⍸'.'=nsname,'.')↑¨nsname ⍝ growing paths
                          :If 9.1≠⎕NC⊂name ⍝ Not a namespace
                              name ⎕NS''
                          :EndIf
                      :EndFor
                  :Else
                      →0
                  :EndTrap
              :EndIf
          :EndIf
          r←⍎nsname
        ∇

    :EndNamespace
    :Namespace Test
⍝ Put the ]link UCMD and FileSystemWatcher through it's paces
⍝ Call Run with a right argument containing a folder name which can be used for the test
⍝ For example:
⍝   Run 'c:\tmp\linktest'

        ⎕IO←1 ⋄ ⎕ML←1

        ∇ r←Run folder;name;foo;ns;nil;ac;bc;tn;goo;old;new;U;link;file;cb;z;zzz;olddd;zoo;goofile;t;m
         
          :If 'Windows'≢7↑⊃'.'⎕WG'APLVersion'
              r←'Unable to run tests - Microsoft Windows is required to test the FileSystemWatcher'
              →0
          :EndIf
         
          U←##.Utils
         
          {}⎕SE.UCMD'udebug on'
          ⎕SE.Link.DEBUG←0 ⍝ 1 = Trace, 2 = Stop on entry
         
          :If 0≠⎕NC'⎕SE.Link.Links'
          :AndIf 0≠≢⎕SE.Link.Links
              Log'Please reset all links and try again.'
              Log ⎕SE.UCMD'link'
              →0
          :EndIf
         
          folder←∊1 ⎕NPARTS folder,(0=≢folder)/'/temp/linktest' ⍝ Normalise
          name←2⊃⎕NPARTS folder
         
          ⎕MKDIR folder ⍝ 2 ⎕NDELETE folder
          #.⎕EX name
         
          cb←' -onRead=⎕SE.Link.Test.onRead -onWrite=⎕SE.Link.Test.onWrite'
          ⎕SE.UCMD'link #.',name,' ',folder,cb
          assert'1=≢⎕SE.Link.Links'
          link←⊃⎕SE.Link.Links
          ns←#⍎name
         
      ⍝ Create a monadic function
          (⊂foo←' r←foo x' ' x x')⎕NPUT folder,'/foo.dyalog'
          assert'foo≡ns.⎕NR ''foo'''
      ⍝ Create a niladic / non-explicit function
          (⊂nil←' nil' ' 2+2')⎕NPUT folder,'/nil.dyalog'
          assert'nil≡ns.⎕NR ''nil'''
         
      ⍝ Create sub-folder
          ⎕MKDIR folder,'/sub'
          assert'9.1=ns.⎕NC ⊂''sub'''
         
      ⍝ Put a copy of foo in the folder
          (⊂foo)⎕NPUT folder,'/sub/foo.dyalog'
          assert'foo≡ns.sub.⎕NR ''foo'''
         
      ⍝ Create a class with missing dependency
          (⊂ac←':Class aClass : bClass' ':EndClass')⎕NPUT folder,'/sub/aClass.dyalog'
          assert'9=ns.sub.⎕NC ''aClass'''
          assert'ac≡⎕SRC ns.sub.aClass'
         
      ⍝ Now add the base class
          (⊂bc←':Class bClass' ':EndClass')⎕NPUT folder,'/sub/bClass.dyalog'
          assert'9=ns.sub.⎕NC ''bClass'''
          assert'bc≡⎕SRC ns.sub.bClass'
         
      ⍝ Check that the derived class works
          :Trap 0 ⋄ {}⎕NEW ns.sub.aClass
          :Else ⋄ ⎕←'NB: Unable to instantiate aClass: ',⊃⎕DM
          :EndTrap
         
      ⍝ Rename the sub-folder
          (folder,'/bus')⎕NMOVE folder,'/sub'
          assert'9.1=ns.⎕NC ⊂''bus'''              ⍝ bus is a namespace
          assert'3=ns.bus.⎕NC ''foo'''             ⍝ bus.foo is a function
          :If ~∨/'/bus/foo.dyalog'⍷4⊃U.GetLinkInfo ns.bus'foo'
              ⎕←'*** NB https://github.com/mkromberg/link/issues/2 still not resolved'
          :EndIf
          assert'0=ns.⎕NC ''sub'''                 ⍝ sub is gone
         
      ⍝ Now copy a file containing a function
          old←U.GetLinkInfo ns'foo'
          (folder,'/foo - copy.dyalog')⎕NCOPY folder,'/foo.dyalog' ⍝ simulate copy/paste
          ⎕DL 1 ⍝ Allow FileSystemWatcher time to react
          goofile←folder,'/goo.dyalog'
          goofile ⎕NMOVE folder,'/foo - copy.dyalog' ⍝ followed by rename
          ⎕DL 1 ⍝ Allow FileSystemWatcher some time to react
      ⍝ Verify that the old function has NOT become linked to the new file
          assert'old≡new←U.GetLinkInfo ns''foo'''
         
      ⍝ Now edit the new file so it "accidentally" defines 'zoo'
          tn←goofile ⎕NTIE 0
          'z'⎕NREPLACE tn 5 80 ⍝ (beware UTF-8 encoded file)
          ⎕NUNTIE tn
      ⍝ Validate that this did cause goo to arrive in the workspace
          zoo←' r←zoo x' ' x x'
          assert'zoo≡ns.⎕NR ''zoo'''
         
      ⍝ Now edit the new file so it finally defines 'goo'
          tn←goofile ⎕NTIE 0
          'g'⎕NREPLACE tn 5 80 ⍝ (beware UTF-8 encoded file)
          ⎕NUNTIE tn
      ⍝ Validate that this did cause goo to arrive in the workspace
          goo←' r←goo x' ' x x'
          assert'goo≡ns.⎕NR ''goo'''
      ⍝ Also validate that zoo is now gone
          assert'0=ns.⎕NC ''zoo'''
         
      ⍝ Now simulate changing goo using the editor and verify the file is updated
          ns'goo'⎕SE.Link.Fix' r←goo x' ' r←x x x'
          assert'(ns.⎕NR ''goo'')≡⊃⎕NGET goofile 1'
         
          ⎕SE.Link.Expunge'ns.goo' ⍝ Test "expunge"
          assert'0=⎕NEXISTS goofile'
         
      ⍝ Now test the Notify function - and verify the System Variable setting trick
         
          link.fsw.Object.EnableRaisingEvents←0 ⍝ Disable notifications
          (⊂':Namespace _SV' '##.(⎕IO←0)' ':EndNamespace')⎕NPUT file←folder,'/bus/_SV.dyalog'
          ⎕SE.Link.Notify'created'file
         
          assert'0=ns.bus.⎕IO'
          assert'1=ns.⎕IO'
         
          link.fsw.Object.EnableRaisingEvents←1 ⍝ Re-enable watcher
         
      ⍝ Now test whether exits implement ".charmat" support
      ⍝ First, write vars in the workspace to file'
         
          ns.cm←↑ns.cv←'Line 1' 'Line two'
          ns'cm'⎕SE.Link.Fix ⍬ ⍝ Inform it charmat was edited
          ns'cv'⎕SE.Link.Fix ⍬ ⍝ Ditto for charvec
          assert'ns.cm≡↑⊃⎕NGET (folder,''/cm.charmat'') 1'
          assert'ns.cv≡⊃⎕NGET (folder,''/cv.charvec'') 1'
         
      ⍝ Then verify that modifying the file brings changes back
          (⊂ns.cv←ns.cv,⊂'Line three')⎕NPUT(folder,'/cv.charvec')1
          (⊂↓ns.cm←↑ns.cv)⎕NPUT(folder,'/cm.charmat')1
         
          assert'ns.cm≡↑⊃⎕NGET (folder,''/cm.charmat'') 1'
          assert'ns.cv≡⊃⎕NGET (folder,''/cv.charvec'') 1'
         
      ⍝ Now tear it all down again:
      ⍝ First the sub-folder
         
          2 ⎕NDELETE folder,'/bus'
          assert'0=⎕NC ''ns.bus'''
         
      ⍝ The variables
          ⎕NDELETE folder,'/cv.charvec'
          ⎕NDELETE folder,'/cm.charmat'
          assert'0 0≡ns.⎕NC 2 2⍴''cmcv'''
         
      ⍝ The the functions, one by one
          ⎕NDELETE folder,'/nil.dyalog'
          assert'0=ns.⎕NC ''nil'''
          ⎕NDELETE folder,'/foo.dyalog'
          assert'0=≢ns.⎕NL -⍳10' ⍝ top level namespace is now empty
         
         EXIT: ⍝ →EXIT to aborted test and clean up
          ⎕SE.Link.DEBUG←0
          ⎕SE.UCMD']link #.',name,' -reset'
          assert'0=≢⎕SE.Link.Links'
         
          z←⊃¨5176⌶⍬ ⍝ Check all links have been cleared
          :If ∨/m←((≢folder)↑¨z)∊⊂folder
              ⎕←'*** Links not cleared:'
              ⎕←⍪m/z
          :EndIf
         
          2 ⎕NDELETE folder    ⍝
          assert'9=#.⎕NC name' ⍝ After ]link -reset this should not remove the namespace
          #.⎕EX name
         
          Log'Tests passed OK'
        ∇

        ∇ Log x
          ⎕←x ⍝ This might get more sophisticated someday
        ∇

        ∇ assert expr;maxwait;end;timeout
      ⍝ Asynchronous assert: We don't know how quickly the FileSystemWatcher will do something
          end←3000+3⊃⎕AI ⍝ 3s
          timeout←0
         
          :While 0∊⍎expr
              ⎕DL 0.1
          :Until timeout←end<3⊃⎕AI
         
          'assertion failed'⎕SIGNAL timeout/11
        ∇

   ⍝ Callback functions to implement .charmat & .charvec support

        ∇ r←onRead(type file nsname);⎕TRAP;parts;data;extn
          r←1 ⍝ Carry on, Soldier!
          :If (⊂extn←3⊃parts←⎕NPARTS file)∊'.charmat' '.charvec'
              :Select type
              :Case 'deleted'
                  (⍎nsname).⎕EX 2⊃parts
                  r←0
              :CaseList 'changed' 'renamed' 'created'
                  data←(↑⍣(extn≡'.charmat'))⊃⎕NGET file 1
                  ⍎nsname,'.',(2⊃parts),'←data'
                  r←0 ⍝ We're done
              :EndSelect
          :EndIf
        ∇

        ∇ r←onWrite(ns name oldname nc src file);⎕TRAP;extn
          r←1 ⍝ Do as you wish
         
          :If nc=2 ⍝ A variable
         
              :Select ⎕DR src
         
              :CaseList 80 160 320
                  :If 2=⍴⍴src ⋄ src←↓src
                  :Else ⋄ →0
                  :EndIf
                  extn←⊂'.charmat'
         
              :Case 326
                  :If (1≠⍴⍴src)∨(10|⎕DR¨src)∨.≠0 ⋄ →0
                  :EndIf
                  extn←⊂'.charvec'
         
              :EndSelect
         
              (⊂src)⎕NPUT(∊(extn@3)⎕NPARTS file)1
              r←0 ⍝ No further work required
          :EndIf
        ∇

    :EndNamespace
:EndNamespace
